/* Copyright (c) 2016 Boulder Real Time Technologies, Inc. */
/* All rights reserved */

/* This software module is wholly owned by Boulder Real Time 
        Technologies, Inc. Any use of this software module without
        express written permission from Boulder Real Time Technologies,
        Inc. is prohibited. */

#include <math.h>

#include "Pkt.h"
#include "swapbytes.h"

int
cd11_utc2epoch (char *utcstring, double *epoch)

/*
 *	cd11_utc2epoch will decode a Cd-1.1 UTC 20 character time string to a double epoch time.
 *
 *  input arguments:
 *      char *utcstring	- 20 character Cd-1.1 UTC time string
 *		
 *  output arguments:
 *		double *epoch 	- Epoch time.
 *
 *   Return value - 0 if no errors or -1 if an error.
 */

{
	char str[32];

	if (strlen(utcstring) != 20) {
		register_error (0, "cd11_utc2epoch: strlen(%s) must be 20 but is %ld instead.\n", utcstring, strlen(utcstring));
		return (-1);
	}
	strcpy (str, utcstring);
	str[20] = '\0';
	str[7] = ' ';
	if (zstr2epoch (str, epoch) < 0) {
		register_error (0, "cd11_utc2epoch: zstr2epoch(%s) error.\n", str);
		return (-1);
	}

	return (0);
}

int
cd11_epoch2utc (double epoch, char *utcstring)

/*
 *	cd11_epoch2utc will encode an epoch time into a Cd-1.1 UTC 20 character time string.
 *
 *  input arguments:
 *		double epoch 	- Epoch time.
 *		
 *  output arguments:
 *      char *utcstring	- 20 character Cd-1.1 UTC time string. It is the responsibility
 *      				  of the caller to allocate the string properly.
 *
 *   Return value - 0 if no errors or -1 if an error.
 */

{
	int year, day, hour, minute, milliseconds;
	double second;

	e2h (epoch, &year, &day, &hour, &minute, &second);
	milliseconds = second*1000.0 + 0.5;
	second = milliseconds * 0.001;
	while (second >= 60.0) {
		minute++;
		second -= 60.0;
	}
	while (minute >= 60) {
		hour++;
		minute -= 60;
	}
	while (hour >= 24) {
		day++;
		hour -= 24;
	}
	if (isleap(year)) {
		if (day > 366) {
			year++;
			day -= 366;
		}
	} else {
		if (day > 365) {
			year++;
			day -= 365;
		}
	}

	sprintf (utcstring, "%4d%3.3d %2.2d:%2.2d:%06.3f", year, day, hour, minute, second);

	return (0);
}

int
cd11_unstuff_frame (char *packet, int nbytes, 
					unsigned char **cd11_frame, int *cd11_frame_nbytes)

/*
 *	cd11_unstuff_frame will pick out the verbatim cd1.1 data frame from a CD11 ORB packet.
 *
 *  input arguments:
 *      char *packet	- pointer to raw ORB packet
 *		int nbytes 		- size of raw ORB packet
 *		
 *  output arguments:
 *		unsigned char **cd11_frame 		- Returns a pointer to the verbatim cd1.1 data
 *										  frame contained within the raw ORB packet.
 *										  Note that this is not an allocated buffer but
 *										  a pointer into the packet array.
 *										  The cd1.1 data packet starts with the 
 *										  "number of channels" field as define on page 17
 *										  of "Formats and Protocols for Continuous
 *										  Data CD-1.1", Table 2.9, immediately after
 *										  the frame header and goes through to end end
 *										  of the channel sub-frame blocks immediately
 *										  before the frame trailer (i.e. the returned
 *										  frame does not include the frame header and frame 
 *										  trailer bytes).
 *		int *cd11_frame_nbytes			- Returned size of verbatim cd1.1 data frame.
 *
 *   Return value - 0 if no errors or -1 if an error.
 */

{
	if (nbytes < 8) {
		register_error (0, "cd11_unstuff_frame: nbytes=%d too small.\n", nbytes);
		return (-1);
	}
	unsigned char *upacket = (unsigned char *) packet;
	unsigned char *ptr = upacket;
	int version;
	mi2hi (&ptr, &version, 1);
	if (version != 0) {
		register_error (0, "cd11_unstuff_frame: version=%d not supported.\n", version);
		return (-1);
	}
	int nchannels;
	mi2hi (&ptr, &nchannels, 1);
	if (nchannels < 1) {
		register_error (0, "cd11_unstuff_frame: nchannels=%d must be > 0.\n", nchannels);
		return (-1);
	}

	int offset_to_frame = 8 + nchannels*48;
	int offset_to_first_subchannel_frame = 32 + nchannels*10;
	offset_to_first_subchannel_frame += offset_to_first_subchannel_frame%4;
	offset_to_first_subchannel_frame += offset_to_frame;
	unsigned char *ptr_frame = upacket + offset_to_frame;
	ptr = ptr_frame;
	int nchannels_frame;
	mi2hi (&ptr, &nchannels_frame, 1);
	if (nchannels != nchannels_frame) {
		register_error (0, "cd11_unstuff_frame: nchannels=%d and nchannels_frame=%d must be the same.\n", nchannels, nchannels_frame);
		return (-1);
	}
	ptr = upacket + offset_to_first_subchannel_frame;
	for (int i=0; i<nchannels; i++) {
		if (ptr - upacket > nbytes) {
			register_error (0, "cd11_unstuff_frame: nbytes=%d too small.\n", nbytes);
			return (-1);
		}
		int size;
		mi2hi (&ptr, &size, 1);
		ptr += size;
	}

	*cd11_frame_nbytes = ptr - ptr_frame;
	*cd11_frame = ptr_frame;

	return (0);
}

Tbl *
get_cd11_channels (unsigned char *cd11_frame)

/*
 *	get_cd11_channels will return a list of Cd-1.1 data frame "site channel[ location]"
 *	codes contained within the frame.
 *
 *  input arguments:
 *		unsigned char *cd11_frame 		- A pointer to the verbatim cd1.1 data
 *										  frame starting with the 
 *										  "number of channels" field as define on page 17
 *										  of "Formats and Protocols for Continuous
 *										  Data CD-1.1", Table 2.9, immediately after
 *										  the frame header and goes through to the end
 *										  of the channel sub-frame blocks immediately
 *										  before the frame trailer (i.e. this
 *										  frame does not include the frame header and frame 
 *										  trailer bytes).
 *
 *   Return value - A newly allocated Tbl with each entry a string in the form
 *   				"<site> <channel>[ <location>]", where <site> is the Cd-1.1 site
 *   				code, <channel> is the Cd-1.1 channel code and <location> is
 *   				the Cd-1.1 location code, or NULL if there is an error. There should
 *   				always be <site> and <lchannel> codes, but <location> codes
 *   				are optional.
 */

{
	Tbl *tbl=NULL;

	unsigned char *ptr = cd11_frame;
	int nchannels;
	mi2hi (&ptr, &nchannels, 1);
	if (nchannels < 1) {
		register_error (0, "get_cd11_channels: nchannels=%d must be > 0.\n", nchannels);
		return (NULL);
	}
	tbl = newtbl (nchannels);
	if (tbl == NULL) {
		register_error (0, "get_cd11_channels: newtbl(%d) error.\n", nchannels);
		return (NULL);
	}
	int offset_to_channel_string = 32;
	ptr = cd11_frame + offset_to_channel_string;
	for (int i=0; i<nchannels; i++) {
		char site[8];
		memset (site, 0, 8);
		strncpy (site, (char *)ptr, 5);
		ptr += 5;
		char channel[8];
		memset (channel, 0, 8);
		strncpy (channel, (char *)ptr, 3);
		ptr += 3;
		char location[8];
		memset (location, 0, 8);
		strncpy (location, (char *)ptr, 2);
		ptr += 2;
		if (!strcmp(site, "")) {
			register_error (0, "get_cd11_channels: NULL site code for channel %d.\n", i);
			freetbl (tbl, free);
			return (NULL);
		}
		if (!strcmp(channel, "")) {
			register_error (0, "get_cd11_channels: NULL channel code for channel %d.\n", i);
			freetbl (tbl, free);
			return (NULL);
		}
		char line[16];
		if (!strcmp(location, "")) {
			sprintf (line, "%s %s", site, channel);
		} else {
			sprintf (line, "%s %s %s", site, channel, location);
		}
		settbl (tbl, -1, strdup(line));
	}

	return (tbl);
}

static void
split_netstachanloc (char *netstachanloc, char *net, char *sta, char *chan, char *loc)

{
        char *ptr;
        int i;

        *net = '\0';
        *sta = '\0';
        *chan = '\0';
        *loc = '\0';
        ptr = net;
        for (i=0; i<strlen(netstachanloc); i++) if (netstachanloc[i] == '_') break; else {*ptr = netstachanloc[i]; ptr++;}
        *ptr = '\0';
        ptr = sta;
        for (i++; i<strlen(netstachanloc); i++) if (netstachanloc[i] == '_') break; else {*ptr = netstachanloc[i]; ptr++;}
        *ptr = '\0';
        ptr = chan;
        for (i++; i<strlen(netstachanloc); i++) if (netstachanloc[i] == '_') break; else {*ptr = netstachanloc[i]; ptr++;}
        *ptr = '\0';
        ptr = loc;
        for (i++; i<strlen(netstachanloc); i++) if (netstachanloc[i] == '_') break; else {*ptr = netstachanloc[i]; ptr++;}
        *ptr = '\0';
}

int
cd11_stuff (unsigned char *cd11_frame, int cd11_frame_nbytes, Tbl *seed_codes, 
				char **packet, int *nbytes, int *packetsz)

/*
 *	cd11_stuff will stuff a CD11 ORB packet from cd1.1 data frame.
 *
 *  input arguments:
 *		unsigned char *cd11_frame 		- A pointer to the verbatim cd1.1 data
 *										  frame starting with the 
 *										  "number of channels" field as define on page 17
 *										  of "Formats and Protocols for Continuous
 *										  Data CD-1.1", Table 2.9, immediately after
 *										  the frame header and goes through to the end
 *										  of the channel sub-frame blocks immediately
 *										  before the frame trailer (i.e. this
 *										  frame does not include the frame header and frame 
 *										  trailer bytes).
 *		int cd11_frame_nbytes			- Size of cd11_frame.
 *		Tbl *seed_codes					- This is a table with string elements defining
 *										  the SEED net_sta_chan[_loc] codes for each channel
 *										  in the CD-1.1 data frame in the form
 *										  "net_sta_chan[_loc][ segtype]". Site, channel and 
 *										  location codes are contained in the CD-1.1 data frame.
 *										  However there are no network codes in the CD-1.1
 *										  data frame and the names for site, channel and
 *										  location may not conform to SEED standards.
 *										  This table provides a translation between the
 *										  CD-1.1 names and the SEED names. 
 *										  All of the metadata values needed to
 *										  define an ORB waveform packet (calib, calper,
 *										  samprate, etc.) can be determined from the CD-1.1
 *										  data frame other than segtype. A segtype value
 *										  can be optionally added after the net_sta_chan[_loc] 
 *										  string, white space delimited. If segtype is
 *										  not specified, then NULL segtype (-) values
 *										  will be used instead.
 *										  The order of the
 *										  names in this table must match the order of the
 *										  channel sub-frames in the CD-1.1 data frame.
 *										  If this is NULL or there are not enough entries
 *										  for each channel, then an error will occur.
 *		
 *  output arguments:
 *      char **packet
 *      int *nbytes
 *      int *packetsz					- Returned ORB packet ready for orbput. These are
 *										  the same arguments as in stuffPkt(3)
 *
 *   Return value - 0 if no errors or -1 if an error.
 */

{
	if (seed_codes == NULL) {
		register_error (0, "cd11_stuff: seed_codes table cannot be NULL.\n");
		return (-1);
	}

	unsigned char *ptr = cd11_frame;
	int nchannels;
	mi2hi (&ptr, &nchannels, 1);
	if (nchannels < 1) {
		register_error (0, "cd11_stuff: nchannels=%d must be > 0.\n", nchannels);
		return (-1);
	}
	if (maxtbl(seed_codes) < nchannels) {
		register_error (0, "cd11_stuff: seed_codes table too small.\n");
		return (-1);
	}

	int size = cd11_frame_nbytes + 8 + nchannels*48;
	if (*packet == NULL) {
		*packet = (char *) malloc (size);
		if (*packet == NULL) {
			register_error (0, "cd11_stuff: malloc(packet,%d) error.\n", size);
			return (-1);
		}
		memset (*packet, 0, size);
		*packetsz = size;
	} else if (size > *packetsz) {
		*packet = (char *) realloc (*packet, size);
		if (*packet == NULL) {
			register_error (0, "cd11_stuff: realloc(packet,%d) error.\n", size);
			return (-1);
		}
		memset (*packet, 0, size);
		*packetsz = size;
	}

	ptr = (unsigned char *) *packet;
	int version = 0;
	hi2mi (&version, &ptr, 1);
	hi2mi (&nchannels, &ptr, 1);

	for (int i=0; i<nchannels; i++) {
		char *line = (char *) gettbl (seed_codes, i);
		char netstachanloc[64], segtype[4];
		int ret = sscanf (line, "%s %s", netstachanloc, segtype);
		if (ret < 0) {
			register_error (1, "cd11_stuff: sscanf(seed_codes) error.\n");
			return (-1);
		}
		if (ret < 1) {
			register_error (1, "cd11_stuff: sscanf(seed_codes,%s) error - no netstachanloc.\n", line);
			return (-1);
		}
		if (ret < 2) strcpy (segtype, "-");
		char net[64], sta[64], chan[64], loc[64];
		split_netstachanloc (netstachanloc, net, sta, chan, loc);
		if (!strcmp(sta, "")) {
			register_error (1, "cd11_stuff: No sta code in %s.\n", netstachanloc);
			return (-1);
		}
		if (!strcmp(chan, "")) {
			register_error (1, "cd11_stuff: No chan code in %s.\n", netstachanloc);
			return (-1);
		}
		memset (ptr, 0, 8);
		strncpy ((char *)ptr, net, 7);
		ptr += 8;
		memset (ptr, 0, 16);
		strncpy ((char *)ptr, sta, 15);
		ptr += 16;
		memset (ptr, 0, 16);
		strncpy ((char *)ptr, chan, 15);
		ptr += 16;
		memset (ptr, 0, 4);
		if (strcmp(loc, "")) strncpy ((char *)ptr, loc, 2);
		ptr += 4;
		memset (ptr, 0, 4);
		strncpy ((char *)ptr, segtype, 2);
		ptr += 4;
	}

	memcpy (ptr, cd11_frame, cd11_frame_nbytes);
	*nbytes = size;

	return (0);
}

static void
myprint (FILE *file, char *fmt, ...)

{
	va_list args;

	va_start (args, fmt);

	if (file) {
		vfprintf (file, fmt, args);
	} else {
		char string[512];
		vsprintf (string, fmt, args);
		elog_notify (0, "%s", string);
	}
}

void
cd11s_show (FILE *file, unsigned char *cd11s_frame, char *prefix) 

/*
 *	cd11_show will print the contents of a Cd-1.1 sub-channel frame.
 *
 *  input arguments:
 *		FILE *file 			  - An open file pointer for printing. If this is NULL, then
 *						  		elog_notify calls will be made to print the results.
 *		unsigned char *cd11s_frame - A pointer to the verbatim Cd-1.1 sub-channel frame.
 *		char *prefix 		  - A string that will prefix all of the print lines.
 */

{
	unsigned char buf[512];
	int i, j, n;
	unsigned char *ptr;
	char string[512];
	char string2[512];
	float f;

	myprint (file, "%s: New CD11S packet\n", prefix);
	memcpy (buf, cd11s_frame, 4);
	cd11s_frame += 4;
	ptr = buf;
	mi2hi (&ptr, &i, 1);
	myprint (file, "%s: channel_length       : %2.2x %2.2x %2.2x %2.2x (%d)\n", prefix,
					buf[0], buf[1], buf[2], buf[3], i);
	memcpy (buf, cd11s_frame, 4);
	cd11s_frame += 4;
	ptr = buf;
	mi2hi (&ptr, &i, 1);
	myprint (file, "%s: authentication_offset: %2.2x %2.2x %2.2x %2.2x (%d)\n", prefix,
					buf[0], buf[1], buf[2], buf[3], i);
	memcpy (buf, cd11s_frame, 24);
	cd11s_frame += 24;
	strcpy (string, "");
	for (i=0; i<24; i++) {
		char number[4];
		sprintf (number, "%2.2x", buf[i]);
		if (strcmp(string, "")) strcat (string, " ");
		strcat (string, number);
	}
	myprint (file, "%s: channel_description  : %s\n", prefix, string);
	myprint (file, "%s:    authentication    : %d\n", prefix, buf[0]);
	myprint (file, "%s:    compression       : %d\n", prefix, buf[1]);
	myprint (file, "%s:    sensor_type       : %d\n", prefix, buf[2]);
	myprint (file, "%s:    option_flag       : %d\n", prefix, buf[3]);
	strncpy (string, (char *)&(buf[4]), 5);
	string[5] = '\0';
	myprint (file, "%s:    site_name         : %s\n", prefix, string);
	strncpy (string, (char *)&(buf[9]), 3);
	string[3] = '\0';
	myprint (file, "%s:    channel_name      : %s\n", prefix, string);
	strncpy (string, (char *)&(buf[12]), 2);
	string[2] = '\0';
	myprint (file, "%s:    location_name     : %s\n", prefix, string);
	strncpy (string, (char *)&(buf[14]), 2);
	string[2] = '\0';
	myprint (file, "%s:    data_format       : %s\n", prefix, string);
	ptr = &(buf[16]);
	mf2hf (&ptr, &f, 1);
	myprint (file, "%s:    calib             : %f\n", prefix, f);
	mf2hf (&ptr, &f, 1);
	myprint (file, "%s:    calper            : %f\n", prefix, f);
	memcpy (buf, cd11s_frame, 20);
	strcpy (string, "");
	for (i=0; i<20; i++) {
		char number[4];
		sprintf (number, "%2.2x", buf[i]);
		if (strcmp(string, "")) strcat (string, " ");
		strcat (string, number);
	}
	strncpy (string2, (char *)cd11s_frame, 20);
	string2[20] = '\0';
	cd11s_frame += 20;
	myprint (file, "%s: time_stamp           : %s (%s)\n", prefix, string, string2);
	memcpy (buf, cd11s_frame, 4);
	cd11s_frame += 4;
	ptr = buf;
	mi2hi (&ptr, &i, 1);
	myprint (file, "%s: subframe_time_length : %2.2x %2.2x %2.2x %2.2x (%d)\n", prefix,
					buf[0], buf[1], buf[2], buf[3], i);
	memcpy (buf, cd11s_frame, 4);
	cd11s_frame += 4;
	ptr = buf;
	mi2hi (&ptr, &i, 1);
	myprint (file, "%s: samples              : %2.2x %2.2x %2.2x %2.2x (%d)\n", prefix,
					buf[0], buf[1], buf[2], buf[3], i);
	memcpy (buf, cd11s_frame, 4);
	cd11s_frame += 4;
	ptr = buf;
	mi2hi (&ptr, &n, 1);
	myprint (file, "%s: channel_status_size  : %2.2x %2.2x %2.2x %2.2x (%d)\n", prefix,
					buf[0], buf[1], buf[2], buf[3], n);
	memcpy (buf, cd11s_frame, n);
	cd11s_frame += n;
	strcpy (string, "");
	for (i=0; i<n; i++) {
		char number[4];
		sprintf (number, "%2.2x", buf[i]);
		if (strcmp(string, "")) strcat (string, " ");
		strcat (string, number);
	}
	myprint (file, "%s: channel_status_data  : %s\n", prefix, string);
	myprint (file, "%s:    format            : %d\n", prefix, buf[0]);
	myprint (file, "%s:    data_status       : 0x%2.2x\n", prefix, buf[1]);
	myprint (file, "%s:    channel_security  : 0x%2.2x\n", prefix, buf[2]);
	myprint (file, "%s:    miscellaneous_status: 0x%2.2x\n", prefix, buf[3]);
	myprint (file, "%s:    voltage_indicator : 0x%2.2x\n", prefix, buf[4]);
	strcpy (string, "");
	for (i=0; i<20; i++) {
		char number[4];
		sprintf (number, "%2.2x", buf[i+8]);
		if (strcmp(string, "")) strcat (string, " ");
		strcat (string, number);
	}
	strncpy (string2, (char *)&(buf[8]), 20);
	string2[20] = '\0';
	myprint (file, "%s:    last_gps_time     : %s (%s)\n", prefix, string, string2);
	ptr = &(buf[28]);
	mi2hi (&ptr, &i, 1);
	myprint (file, "%s:    clock_drift       : %2.2x %2.2x %2.2x %2.2x (%d)\n", prefix,
					buf[28], buf[29], buf[30], buf[31], i);
	memcpy (buf, cd11s_frame, 4);
	cd11s_frame += 4;
	ptr = buf;
	mi2hi (&ptr, &n, 1);
	myprint (file, "%s: data_size            : %2.2x %2.2x %2.2x %2.2x (%d)\n", prefix,
					buf[0], buf[1], buf[2], buf[3], n);
	int first = 1;
	for (i=0,j=0; i<n; i++,j++) {
		char number[4];
		sprintf (number, "%2.2x", cd11s_frame[i]);
		if (j == 0) strcpy (string, "");
		if (strcmp(string, "")) strcat (string, " ");
		strcat (string, number);
		if (j == 31) {
			j = -1;
			if (first) myprint (file, "%s: channel_data         : %s\n", prefix, string);
			else myprint (file, "%s:                        %s\n", prefix, string);
			first = 0;
		}
	}
	if (j > 0) {
		if (first) myprint (file, "%s: channel_data         : %s\n", prefix, string);
		else myprint (file, "%s:                        %s\n", prefix, string);
	}
	cd11s_frame += (n + n%4);
	memcpy (buf, cd11s_frame, 4);
	cd11s_frame += 4;
	ptr = buf;
	mi2hi (&ptr, &i, 1);
	myprint (file, "%s: subframe_count       : %2.2x %2.2x %2.2x %2.2x (%d)\n", prefix,
					buf[0], buf[1], buf[2], buf[3], i);
	memcpy (buf, cd11s_frame, 4);
	cd11s_frame += 4;
	ptr = buf;
	mi2hi (&ptr, &i, 1);
	myprint (file, "%s: authentication_key   : %2.2x %2.2x %2.2x %2.2x (%d)\n", prefix,
					buf[0], buf[1], buf[2], buf[3], i);
	memcpy (buf, cd11s_frame, 4);
	cd11s_frame += 4;
	ptr = buf;
	mi2hi (&ptr, &n, 1);
	myprint (file, "%s: authentication_size  : %2.2x %2.2x %2.2x %2.2x (%d)\n", prefix,
					buf[0], buf[1], buf[2], buf[3], n);
	first = 1;
	for (i=0,j=0; i<n; i++,j++) {
		char number[4];
		sprintf (number, "%2.2x", cd11s_frame[i]);
		if (j == 0) strcpy (string, "");
		if (strcmp(string, "")) strcat (string, " ");
		strcat (string, number);
		if (j == 31) {
			j = -1;
			if (first) myprint (file, "%s: authentication_value : %s\n", prefix, string);
			else myprint (file, "%s:                        %s\n", prefix, string);
			first = 0;
		}
	}
	if (j > 0) {
		if (first) myprint (file, "%s: authentication_value : %s\n", prefix, string);
		else myprint (file, "%s:                        %s\n", prefix, string);
	}
}

void
cd11_show (FILE *file, unsigned char *cd11_frame, char *prefix) 

/*
 *	cd11_show will print the contents of a Cd-1.1 data frame.
 *
 *  input arguments:
 *		FILE *file 			  - An open file pointer for printing. If this is NULL, then
 *						  		elog_notify calls will be made to print the results.
 *		unsigned char *cd11_frame 		- A pointer to the verbatim cd1.1 data
 *										  frame starting with the 
 *										  "number of channels" field as define on page 17
 *										  of "Formats and Protocols for Continuous
 *										  Data CD-1.1", Table 2.9, immediately after
 *										  the frame header and goes through to the end
 *										  of the channel sub-frame blocks immediately
 *										  before the frame trailer (i.e. this
 *										  frame does not include the frame header and frame 
 *										  trailer bytes).
 *		char *prefix 		  - A string that will prefix all of the print lines.
 */

{
	unsigned char buf[512];
	int i, j;
	unsigned char *ptr;
	char string[512];
	char string2[512];

	myprint (file, "%s: New CD11 packet\n", prefix);
	memcpy (buf, cd11_frame, 4);
	cd11_frame += 4;
	ptr = buf;
	mi2hi (&ptr, &i, 1);
	int number_of_channels = i;
	myprint (file, "%s: number_of_channels   : %2.2x %2.2x %2.2x %2.2x (%d)\n", prefix,
					buf[0], buf[1], buf[2], buf[3], i);
	memcpy (buf, cd11_frame, 4);
	cd11_frame += 4;
	ptr = buf;
	mi2hi (&ptr, &i, 1);
	myprint (file, "%s: frame_time_length    : %2.2x %2.2x %2.2x %2.2x (%d)\n", prefix,
					buf[0], buf[1], buf[2], buf[3], i);
	memcpy (buf, cd11_frame, 20);
	strcpy (string, "");
	for (i=0; i<20; i++) {
		char number[4];
		sprintf (number, "%2.2x", buf[i]);
		if (strcmp(string, "")) strcat (string, " ");
		strcat (string, number);
	}
	strncpy (string2, (char *)cd11_frame, 20);
	string2[20] = '\0';
	cd11_frame += 20;
	myprint (file, "%s: nominal_time      : %s (%s)\n", prefix, string, string2);
	memcpy (buf, cd11_frame, 4);
	cd11_frame += 4;
	ptr = buf;
	mi2hi (&ptr, &i, 1);
	int channel_string_count = i;
	channel_string_count += channel_string_count%4;
	myprint (file, "%s: channel_string_count : %2.2x %2.2x %2.2x %2.2x (%d)\n", prefix,
					buf[0], buf[1], buf[2], buf[3], i);
	int first = 1;
	char site[8], channel[8], location[8];
	memset (site, 0, 8);
	memset (channel, 0, 8);
	memset (location, 0, 8);
	for (i=0,j=0; i<channel_string_count; i++,j++) {
		char number[4];
		sprintf (number, "%2.2x", cd11_frame[i]);
		if (j == 0) strcpy (string, "");
		if (strcmp(string, "")) strcat (string, " ");
		strcat (string, number);
		if (j < 5) {
			site[j] = cd11_frame[i];
		} else if (j < 8) {
			channel[j-5] = cd11_frame[i];
		} else {
			location[j-8] = cd11_frame[i];
		}
		if (j == 9) {
			j = -1;
			if (first) myprint (file, "%s: channel_string       : %s ('%s', '%s', '%s')\n", prefix, string,
																	site, channel, location);
			else myprint (file, "%s:                        %s ('%s', '%s', '%s')\n", prefix, string,
																	site, channel, location);
			first = 0;
			memset (site, 0, 8);
			memset (channel, 0, 8);
			memset (location, 0, 8);
		}
	}
	if (j > 0) {
		if (first) myprint (file, "%s: channel_string       : %s ('%s', '%s', '%s')\n", prefix, string,
																site, channel, location);
		else myprint (file, "%s:                        %s ('%s', '%s', '%s')\n", prefix, string,
																site, channel, location);
	}
	cd11_frame += channel_string_count;
	for (int i=0; i<number_of_channels; i++) {
		myprint (file, "%s: sub-channel frame    : %d\n", prefix, i);
		cd11s_show (file, cd11_frame, prefix);
		memcpy (buf, cd11_frame, 4);
		cd11_frame += 4;
		ptr = buf;
		mi2hi (&ptr, &j, 1);
		cd11_frame += j;
	}
}